// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Text;

using Smdn.Formats.Mime.Decode;
using Smdn.Formats.Mime.Encode;

namespace Smdn.Formats.Mime {
  public static class ContentStreamUtils {
    public static Stream CreateDecodingStream(Stream contentStream, string transferEncoding)
    {
      if (transferEncoding == null)
        throw new ArgumentNullException("transferEncoding");

      return CreateDecodingStream(contentStream, Parser.ToTransferEncoding(transferEncoding));
    }

    public static Stream CreateDecodingStream(Stream contentStream, MimeMessageTransferEncoding transferEncoding)
    {
      if (contentStream == null)
        throw new ArgumentNullException("contentStream");

      switch (transferEncoding) {
        case MimeMessageTransferEncoding.SevenBit:
        case MimeMessageTransferEncoding.EightBit:
        case MimeMessageTransferEncoding.Binary:
          return contentStream;
        case MimeMessageTransferEncoding.Base64:
          return new Base64ContentDecodingStream(contentStream);
        case MimeMessageTransferEncoding.QuotedPrintable:
          return new QuotedPrintableContentDecodingStream(contentStream);
        case MimeMessageTransferEncoding.UUEncode:
        case MimeMessageTransferEncoding.GZip64:
        default:
          throw new NotSupportedException(string.Format("unsupported transfer encoding: {0}", transferEncoding));
      }
    }

    public static StreamReader CreateTextReader(Stream contentStream, string transferEncoding, string charset)
    {
      return CreateTextReader(contentStream,
                              Parser.ToTransferEncoding(transferEncoding),
                              charset == null ? Charsets.ISO8859_1 : Charsets.FromString(charset));
    }

    public static StreamReader CreateTextReader(Stream contentStream, MimeMessageTransferEncoding transferEncoding, Encoding charset)
    {
      if (transferEncoding == MimeMessageTransferEncoding.Binary)
        throw new InvalidOperationException("can't create TextReader from message of binary transfer encoding");

      return new StreamReader(CreateDecodingStream(contentStream, transferEncoding), charset);
    }

    public static BinaryReader CreateBinaryReader(Stream contentStream, string transferEncoding)
    {
      return CreateBinaryReader(contentStream,
                                Parser.ToTransferEncoding(transferEncoding),
                                null);
    }

    public static BinaryReader CreateBinaryReader(Stream contentStream, MimeMessageTransferEncoding transferEncoding)
    {
      return CreateBinaryReader(contentStream,
                                transferEncoding,
                                null);
    }

    public static BinaryReader CreateBinaryReader(Stream contentStream, MimeMessageTransferEncoding transferEncoding, Encoding charset)
    {
      var stream = CreateDecodingStream(contentStream, transferEncoding);

      if (charset == null)
        return new BinaryReader(stream);
      else
        return new BinaryReader(stream, charset);
    }

    public static Stream CreateEncodingStream(Stream contentStream, string transferEncoding)
    {
      if (transferEncoding == null)
        throw new ArgumentNullException("transferEncoding");

      return CreateEncodingStream(contentStream, Parser.ToTransferEncoding(transferEncoding), MimeFormat.Unspecified);
    }

    public static Stream CreateEncodingStream(Stream contentStream, MimeMessageTransferEncoding transferEncoding)
    {
      return CreateEncodingStream(contentStream, transferEncoding, MimeFormat.Unspecified);
    }

    public static Stream CreateEncodingStream(Stream contentStream, MimeMessageTransferEncoding transferEncoding, MimeFormat format)
    {
      if (contentStream == null)
        throw new ArgumentNullException("contentStream");
      if (format == null)
        throw new ArgumentNullException("format");

      switch (transferEncoding) {
        case MimeMessageTransferEncoding.SevenBit:
        case MimeMessageTransferEncoding.EightBit:
        case MimeMessageTransferEncoding.Binary:
          return contentStream;
        case MimeMessageTransferEncoding.Base64:
          return new Base64ContentEncodingStream(contentStream, format);
        case MimeMessageTransferEncoding.QuotedPrintable:
          return new QuotedPrintableContentEncodingStream(contentStream, format);
        case MimeMessageTransferEncoding.UUEncode:
        case MimeMessageTransferEncoding.GZip64:
        default:
          throw new NotSupportedException(string.Format("unsupported transfer encoding: {0}", transferEncoding));
      }
    }
  }
}
