// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Net.Imap4.Client.Transaction;
using Smdn.Net.Imap4.Client.Transaction.BuiltIn;

namespace Smdn.Net.Imap4.Client.Session {
  partial class ImapSession {
    /*
     * transaction methods : selected state
     */

    /// <summary>sends CHECK command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Check()
    {
      RejectNonSelectedState();

      using (var t = new CheckTransaction(connection)) {
        return ProcessTransaction(t, handlesIncapableAsException);
      }
    }

    /// <summary>sends CLOSE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Close()
    {
      RejectNonAuthenticatedState();

      if (state != ImapSessionState.Selected)
        return new ImapTransactionResult(ImapTransactionResultCode.RequestDone,
                                         "mailbox has already closed or has not selected yet");

      using (var t = new CloseTransaction(connection)) {
        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return t.Result;
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Expunge()
    {
      long[] discard;

      return Expunge(out discard);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Expunge(out long[] expungedMessages)
    {
      using (var t = new ExpungeTransaction(connection, false)) {
        return ExpungeInternal(t, null, out expungedMessages);
      }
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidExpunge(ImapSequenceSet uidSet)
    {
      long[] discard;

      return UidExpunge(uidSet, out discard);
    }

    /// <summary>sends EXPUNGE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidExpunge(ImapSequenceSet uidSet, out long[] expungedMessages)
    {
      if (uidSet == null)
        throw new ArgumentNullException("uidSet");
      if (!uidSet.IsUidSet)
        throw new ArgumentException("sequence set is not UID", "uidSet");
      if (uidSet.IsEmpty)
        throw new ArgumentException("uid set is empty", "uidSet");

      using (var t = new ExpungeTransaction(connection, true)) {
        return ExpungeInternal(t, uidSet, out expungedMessages);
      }
    }

    private ImapTransactionResult ExpungeInternal(ExpungeTransaction t, ImapSequenceSet uidSet, out long[] expungedMessages)
    {
      RejectNonSelectedState();

      expungedMessages = null;

      // sequence set for UID EXPUNGE
      if (uidSet != null)
        t.Request.Arguments["sequence set"] = uidSet.ToString();

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        expungedMessages = t.Result.Value;

      return t.Result;
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Search(ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return Search(searchingCriteria, null, out matched);
    }

    /// <summary>sends SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Search(ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, false, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult ESearch(ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return ESearch(searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult ESearch(ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, false, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, false, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSearch(ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return UidSearch(searchingCriteria, null, out matched);
    }

    /// <summary>sends UID SEARCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSearch(ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, true, null)) {
        return SearchSortInternal(t, null, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult UidESearch(ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return UidESearch(searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends SEARCH command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESEARCH extension.
    /// </remarks>
    public ImapTransactionResult UidESearch(ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SearchTransaction(connection, true, ImapCapability.ESearch)) {
        return ESearchESortInternal(t, true, null, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Sort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return Sort(sortingCriteria, searchingCriteria, null, out matched);
    }

    /// <summary>sends SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Sort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, false, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult ESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return ESort(sortingCriteria, searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult ESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, false, ImapCapability.ESort)) {
        return ESearchESortInternal(t, false, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, out ImapMatchedSequenceSet matched)
    {
      return UidSort(sortingCriteria, searchingCriteria, (Encoding)null, out matched);
    }

    /// <summary>sends UID SORT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, true, null)) {
        return SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, null, out matched);
      }
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult UidESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      return UidESort(sortingCriteria, searchingCriteria, null, resultOptions, out matched);
    }

    /// <summary>sends UID SORT command with RETURN result specifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support ESORT extension.
    /// </remarks>
    public ImapTransactionResult UidESort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      using (var t = new SortTransaction(connection, true, ImapCapability.ESort)) {
        return ESearchESortInternal(t, true, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);
      }
    }

    private ImapTransactionResult ESearchESortInternal(SearchTransactionBase t, bool uid, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      if (resultOptions == null)
        throw new ArgumentNullException("resultOptions");

      /*
       * RFC 5182 - IMAP Extension for Referencing the Last SEARCH Result
       * http://tools.ietf.org/html/rfc5182
       */
      var saveResult = resultOptions.ContainsOneOf(ImapSearchResultOptions.Save);

      // ESEARCH or ESORT capability
      if (t is SearchTransaction) {
        if (!resultOptions.RequiredCapabilities.Contains(ImapCapability.ESearch))
          resultOptions.RequiredCapabilities.Add(ImapCapability.ESearch);
        if (resultOptions.RequiredCapabilities.Contains(ImapCapability.ESort))
          resultOptions.RequiredCapabilities.Remove(ImapCapability.ESort);
      }
      else if (t is SortTransaction) {
        if (!resultOptions.RequiredCapabilities.Contains(ImapCapability.ESort))
          resultOptions.RequiredCapabilities.Add(ImapCapability.ESort);
        if (resultOptions.RequiredCapabilities.Contains(ImapCapability.ESearch))
          resultOptions.RequiredCapabilities.Remove(ImapCapability.ESearch);
      }

      SearchSortInternal(t, sortingCriteria, searchingCriteria, charset, resultOptions, out matched);

      if (saveResult) {
        var saved = t.Result.Succeeded && (t.Result.GetResponseCode(ImapResponseCode.NotSaved) == null);

        if (saved)
          matched = ImapMatchedSequenceSet.CreateSavedResult(matched == null
                                                             ? matched = new ImapMatchedSequenceSet(ImapSequenceSet.CreateSet(uid, new long[] {}))
                                                             : matched);
      }

      return t.Result;
    }

    private ImapTransactionResult SearchSortInternal(SearchTransactionBase t, ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, Encoding charset, ImapSearchResultOptions resultOptions, out ImapMatchedSequenceSet matched)
    {
      if (t is SortTransaction && sortingCriteria == null)
        throw new ArgumentNullException("sortingCriteria");
      if (searchingCriteria == null)
        throw new ArgumentNullException("searchingCriteria");

      RejectNonSelectedState();

      matched = null;

      /*
       * 6.4.4. SEARCH Command
       *    US-ASCII MUST be supported; other [CHARSET]s MAY be supported.
       * 
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.SORT. SORT Command
       *    The US-ASCII and [UTF-8] charsets MUST be implemented.
       *    All other charsets are optional.
       */
      if (charset == null) {
        if (t is SortTransaction)
          charset = Encoding.UTF8;
        else
          charset = Encoding.ASCII;
      }

      var charsetSpecified = searchingCriteria.SetCharset(charset);

      /*
       * RFC 4466 - Collected Extensions to IMAP4 ABNF
       * http://tools.ietf.org/html/rfc4466
       *
       *    search          = "SEARCH" [search-return-opts]
       *                      SP search-program
       *    search-correlator  = SP "(" "TAG" SP tag-string ")"
       *    search-program     = ["CHARSET" SP charset SP]
       *                         search-key *(SP search-key)
       *                         ;; CHARSET argument to SEARCH MUST be
       *                         ;; registered with IANA.
       *    search-return-data = search-modifier-name SP search-return-value
       *                         ;; Note that not every SEARCH return option
       *                         ;; is required to have the corresponding
       *                         ;; ESEARCH return data.
       *    search-return-opts = SP "RETURN" SP "(" [search-return-opt
       *                         *(SP search-return-opt)] ")"
       *    search-return-opt = search-modifier-name [SP search-mod-params]
       *    search-return-value = tagged-ext-val
       *                         ;; Data for the returned search option.
       *                         ;; A single "nz-number"/"number" value
       *                         ;; can be returned as an atom (i.e., without
       *                         ;; quoting).  A sequence-set can be returned
       *                         ;; as an atom as well.
       *    search-modifier-name = tagged-ext-label
       *    search-mod-params = tagged-ext-val
       *                      ;; This non-terminal shows recommended syntax
       *                      ;; for future extensions.
       */

      /* 
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.SORT. SORT Command
       * The charset argument is mandatory (unlike SEARCH)
       */
      // charset specification
      if (t is SortTransaction || charsetSpecified)
        t.Request.Arguments["charset specification"] = charset.WebName;

      // search-return-opt
      if (resultOptions != null)
        t.Request.Arguments["result specifier"] = resultOptions;

      // searching criteria
      t.Request.Arguments["searching criteria"] = searchingCriteria;

        // sort criteria
      if (sortingCriteria != null)
        t.Request.Arguments["sort criteria"] = sortingCriteria;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        matched = t.Result.Value;

      // TODO: BADCHARSET
      // 6.4.4. SEARCH Command
      //       If the server does not support the specified [CHARSET], it MUST
      //       return a tagged NO response (not a BAD).  This response SHOULD
      //       contain the BADCHARSET response code, which MAY list the
      //       [CHARSET]s supported by the server.

      return t.Result;
    }

    internal ImapTransactionResult UidSort(ImapSortCriteria sortingCriteria, ImapSearchCriteria searchingCriteria, string charset, out ImapMatchedSequenceSet matched)
    {
      if (sortingCriteria == null)
        throw new ArgumentNullException("sortingCriteria");
      if (searchingCriteria == null)
        throw new ArgumentNullException("searchingCriteria");

      RejectNonSelectedState();

      matched = null;

      if (charset == null)
        charset = Encoding.UTF8.WebName;

      using (var t = new SortTransaction(connection, true, null)) {
        // charset specification
        t.Request.Arguments["charset specification"] = charset;

        // searching criteria
        t.Request.Arguments["searching criteria"] = searchingCriteria;

        // sort criteria
        t.Request.Arguments["sort criteria"] = sortingCriteria;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
          matched = t.Result.Value;

        return t.Result;
      }
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Thread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, false)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, Encoding.UTF8, out threadList);
      }
    }

    /// <summary>sends THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Thread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, false)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset, out threadList);
      }
    }

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, Encoding.UTF8, out threadList);
      }
    }

    /// <summary>sends UID THREAD command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset, out threadList);
      }
    }

    private ImapTransactionResult ThreadInternal(ThreadTransaction t, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, Encoding charset, out ImapThreadList threadList)
    {
      if (searchingCriteria == null)
        throw new ArgumentNullException("searchingCriteria");
      if (charset == null)
        throw new ArgumentNullException("charset");

      /*
       * http://tools.ietf.org/html/rfc5256
       * BASE.6.4.THREAD. THREAD Command
       *   The US-ASCII and [UTF-8] charsets MUST be implemented.
       */
      if (charset == null)
        charset = Encoding.UTF8;

      searchingCriteria.SetCharset(charset);

      return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset.WebName, out threadList);
    }

    internal ImapTransactionResult UidThread(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, string charset, out ImapThreadList threadList)
    {
      using (var t = new ThreadTransaction(connection, true)) {
        return ThreadInternal(t, threadingAlgorithm, searchingCriteria, charset ?? Encoding.UTF8.WebName, out threadList);
      }
    }

    private ImapTransactionResult ThreadInternal(ThreadTransaction t, ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchingCriteria, string charset, out ImapThreadList threadList)
    {
      if (threadingAlgorithm == null)
        throw new ArgumentNullException("threadingAlgorithm");

      RejectNonSelectedState();

      threadList = null;

      // threading algorithm
      t.Request.Arguments["threading algorithm"] = threadingAlgorithm;

      // charset specification
      t.Request.Arguments["charset specification"] = charset;

      // searching criteria
      t.Request.Arguments["searching criteria"] = searchingCriteria;

      if (ProcessTransaction(t, handlesIncapableAsException).Succeeded)
        threadList = t.Result.Value;

      return t.Result;
    }

    /// <summary>sends FETCH command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Fetch<TMessageAttribute>(ImapSequenceSet sequenceOrUidSet,
                                                          ImapFetchDataItem messageDataItems,
                                                          out TMessageAttribute[] messages)
      where TMessageAttribute : ImapMessageAttributeBase
    {
      return FetchInternal(sequenceOrUidSet, messageDataItems, null, null, out messages);
    }

    /// <summary>sends FETCH command with CHANGEDSINCE FETCH modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult FetchChangedSince<TMessageAttribute>(ImapSequenceSet sequenceOrUidSet,
                                                                      ImapFetchDataItem messageDataItems,
                                                                      ulong modSeq,
                                                                      out TMessageAttribute[] messages)
      where TMessageAttribute : ImapMessageAttributeBase
    {
      var fetchModifiers = new ImapParenthesizedString("CHANGEDSINCE", modSeq.ToString());

      return FetchInternal(sequenceOrUidSet, messageDataItems, fetchModifiers, ImapCapability.CondStore, out messages);
    }

    private ImapTransactionResult FetchInternal<TMessageAttribute>(ImapSequenceSet sequenceOrUidSet,
                                                                   ImapFetchDataItem messageDataItems,
                                                                   ImapParenthesizedString fetchModifiers,
                                                                   ImapCapability fetchModifiersCapabilityRequirement,
                                                                   out TMessageAttribute[] messages)
      where TMessageAttribute : ImapMessageAttributeBase
    {
      RejectNonSelectedState();

      if (sequenceOrUidSet == null)
        throw new ArgumentNullException("sequenceOrUidSet");
      else if (sequenceOrUidSet.IsEmpty)
        throw new ArgumentException("sequence or uid set is empty", "sequenceOrUidSet");

      if (messageDataItems == null)
        throw new ArgumentNullException("messageDataItems");

      messages = null;

      using (var t = new FetchTransaction<TMessageAttribute>(connection, sequenceOrUidSet.IsUidSet, fetchModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    fetch           = "FETCH" SP sequence-set SP ("ALL" / "FULL" /
         *                      "FAST" / fetch-att /
         *                      "(" fetch-att *(SP fetch-att) ")")
         *                      [fetch-modifiers]
         *                      ;; modifies the original IMAP4 FETCH command to
         *                      ;; accept optional modifiers
         *    fetch-modifiers = SP "(" fetch-modifier *(SP fetch-modifier) ")"
         *    fetch-modifier  = fetch-modifier-name [ SP fetch-modif-params ]
         *    fetch-modif-params  = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    fetch-modifier-name = tagged-ext-label
         */
        // sequence set
        t.Request.Arguments["sequence set"] = sequenceOrUidSet.ToString();

        // message data item names or macro
        t.Request.Arguments["message data item names or macro"] = messageDataItems;

        // fetch-modifiers
        if (fetchModifiers != null)
          t.Request.Arguments["fetch modifiers"] = fetchModifiers;

        if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
          messages = t.Result.Value;

          UpdateBaseUrls(messages);
        }

        return t.Result;
      }
    }

    private void UpdateBaseUrls(ImapMessageAttributeBase[] messages)
    {
      if (messages == null)
        return;

      foreach (var message in messages) {
        var staticAttr = message as IImapMessageStaticAttribute;

        if (staticAttr == null)
          continue;

        staticAttr.SetBaseUrl(SelectedMailbox.Url);

        if (staticAttr.BodyStructure != null)
          staticAttr.BodyStructure.SetBaseUrl(staticAttr.Url);
      }
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Store(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem)
    {
      ImapMessageAttribute[] discard;

      return Store(sequenceOrUidSet, storeDataItem, out discard);
    }

    /// <summary>sends STORE command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Store(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, out ImapMessageAttribute[] messageAttributes)
    {
      return StoreInternal(sequenceOrUidSet, storeDataItem, null, null, out messageAttributes);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq)
    {
      ImapMessageAttribute[] discard;
      ImapSequenceSet discard2;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out discard, out discard2);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapMessageAttribute[] messageAttributes)
    {
      ImapSequenceSet discard;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out messageAttributes, out discard);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapSequenceSet failedMessageSet)
    {
      ImapMessageAttribute[] discard;

      return StoreUnchangedSince(sequenceOrUidSet, storeDataItem, modSeq, out discard, out failedMessageSet);
    }

    /// <summary>sends STORE command with UNCHANGEDSINCE STORE modifier</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method will fail if server does not support CONDSTORE extension.
    /// </remarks>
    [CLSCompliant(false)]
    public ImapTransactionResult StoreUnchangedSince(ImapSequenceSet sequenceOrUidSet, ImapStoreDataItem storeDataItem, ulong modSeq, out ImapMessageAttribute[] messageAttributes, out ImapSequenceSet failedMessageSet)
    {
      failedMessageSet = null;
      messageAttributes = null;

      var storeModifiers = new ImapParenthesizedString("UNCHANGEDSINCE", modSeq.ToString());

      var ret = StoreInternal(sequenceOrUidSet, storeDataItem, storeModifiers, ImapCapability.CondStore, out messageAttributes);

      var modified = lastTransactionResult.GetResponseCode(ImapResponseCode.Modified);

      if (modified is ImapTaggedStatusResponse)
        failedMessageSet = ImapResponseTextConverter.FromModified(modified.ResponseText, sequenceOrUidSet.IsUidSet);

      return ret;
    }

    private ImapTransactionResult StoreInternal(ImapSequenceSet sequenceOrUidSet,
                                                ImapStoreDataItem storeDataItem,
                                                ImapParenthesizedString storeModifiers,
                                                ImapCapability storeModifiersCapabilityRequirement,
                                                out ImapMessageAttribute[] messageAttributes)
    {
      RejectNonSelectedState();

      if (sequenceOrUidSet == null)
        throw new ArgumentNullException("sequenceOrUidSet");
      else if (sequenceOrUidSet.IsEmpty)
        throw new ArgumentException("sequence or uid set is empty", "sequenceOrUidSet");

      if (storeDataItem == null)
        throw new ArgumentNullException("storeDataItem");

      messageAttributes = null;

      using (var t = new StoreTransaction(connection, sequenceOrUidSet.IsUidSet, storeModifiersCapabilityRequirement)) {
        /*
         * RFC 4466 - Collected Extensions to IMAP4 ABNF
         * http://tools.ietf.org/html/rfc4466
         *
         *    store           = "STORE" SP sequence-set [store-modifiers]
         *                      SP store-att-flags
         *                      ;; extend [IMAP4] STORE command syntax
         *                      ;; to allow for optional store-modifiers
         *    store-modifiers =  SP "(" store-modifier *(SP store-modifier)
         *                        ")"
         *    store-modifier  = store-modifier-name [SP store-modif-params]
         *    store-modif-params = tagged-ext-val
         *                      ;; This non-terminal shows recommended syntax
         *                      ;; for future extensions.
         *    store-modifier-name = tagged-ext-label
         */
        // message set
        t.Request.Arguments["message set"] = sequenceOrUidSet.ToString();

        // store-modifiers
        if (storeModifiers != null)
          t.Request.Arguments["store modifiers"] = storeModifiers;

        // message data item name
        t.Request.Arguments["message data item name"] = storeDataItem.ItemName;

        // value for message data item
        t.Request.Arguments["value for message data item"] = storeDataItem;

        ProcessTransaction(t, handlesIncapableAsException);

        messageAttributes = t.Result.Value;

        UpdateBaseUrls(messageAttributes);

        return t.Result;
      }
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, ImapMailbox mailbox)
    {
      ValidateMailboxRelationship(mailbox);

      ImapCopiedUidSet discard;
      ImapMailbox discard2;

      return CopyInternal(sequenceOrUidSet, mailbox.Name, false, out discard, out discard2);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, ImapMailbox mailbox, out ImapCopiedUidSet copiedUids)
    {
      ValidateMailboxRelationship(mailbox);

      ImapMailbox discard;

      return CopyInternal(sequenceOrUidSet, mailbox.Name, false, out copiedUids, out discard);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName)
    {
      ImapCopiedUidSet discard;
      ImapMailbox discard2;

      return CopyInternal(sequenceOrUidSet, mailboxName, false, out discard, out discard2);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids)
    {
      ImapMailbox discard;

      return CopyInternal(sequenceOrUidSet, mailboxName, false, out copiedUids, out discard);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapMailbox createdMailbox)
    {
      ImapCopiedUidSet discard;

      return CopyInternal(sequenceOrUidSet, mailboxName, true, out discard, out createdMailbox);
    }

    /// <summary>sends COPY command</summary>
    /// <remarks>
    /// valid in selected state.
    /// this method tries to automatically create the mailbox when server sent [TRYCREATE] response code.
    /// the out parameter <paramref name="copiedUids"/> will be set if the server supports UIDPLUS extension and returns [COPYUID] responce code, otherwise null.
    /// </remarks>
    public ImapTransactionResult Copy(ImapSequenceSet sequenceOrUidSet, string mailboxName, out ImapCopiedUidSet copiedUids, out ImapMailbox createdMailbox)
    {
      return CopyInternal(sequenceOrUidSet, mailboxName, true, out copiedUids, out createdMailbox);
    }

    private ImapTransactionResult CopyInternal(ImapSequenceSet sequenceOrUidSet, string mailboxName, bool tryCreate, out ImapCopiedUidSet copiedUids, out ImapMailbox createdMailbox)
    {
      RejectNonSelectedState();

      if (sequenceOrUidSet == null)
        throw new ArgumentNullException("sequenceOrUidSet");
      else if (sequenceOrUidSet.IsEmpty)
        throw new ArgumentException("sequence or uid set is empty", "sequenceOrUidSet");

      RejectInvalidMailboxNameArgument(mailboxName);

      copiedUids = null;
      createdMailbox = null;

      ImapTransactionResult failedResult = null;

      for (var i = 0; i < 2; i++) {
        var respTryCreate = false;

        using (var t = new CopyTransaction(connection, sequenceOrUidSet.IsUidSet)) {
          // sequence set
          t.Request.Arguments["sequence set"] = sequenceOrUidSet.ToString();

          // mailbox name
          t.Request.Arguments["mailbox name"] = new ImapMailboxNameString(mailboxName);

          if (ProcessTransaction(t, handlesIncapableAsException).Succeeded) {
            copiedUids = t.Result.Value;
            return t.Result;
          }
          else if (!tryCreate) {
            return t.Result;
          }

          failedResult = t.Result;

          // 6.4.7. COPY Command
          //       If the destination mailbox does not exist, a server SHOULD return
          //       an error.  It SHOULD NOT automatically create the mailbox.  Unless
          //       it is certain that the destination mailbox can not be created, the
          //       server MUST send the response code "[TRYCREATE]" as the prefix of
          //       the text of the tagged NO response.  This gives a hint to the
          //       client that it can attempt a CREATE command and retry the COPY if
          //       the CREATE is successful.
          respTryCreate = (t.Result.GetResponseCode(ImapResponseCode.TryCreate) is ImapTaggedStatusResponse);
        }

        // try create
        if (i == 0 && respTryCreate)
          if (Create(mailboxName, out createdMailbox).Failed)
            return failedResult;
      }

      return failedResult;
    }

    /// <summary>sends UNSELECT command</summary>
    /// <remarks>valid in selected state</remarks>
    public ImapTransactionResult Unselect()
    {
      RejectNonSelectedState();

      using (var t = new UnselectTransaction(connection)) {
        var r = ProcessTransaction(t, handlesIncapableAsException);

        if (r.Succeeded) {
          selectedMailbox = null;
          TransitStateTo(ImapSessionState.Authenticated);
        }

        return r;
      }
    }
  }
}
