// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public class ImapMailboxInfo : IImapUrl {
    public Uri Url {
      get { return mailbox.Url; }
    }

    public string FullName {
      get { return mailbox.Name; }
    }

    public string Name {
      get { return mailbox.LeafName; }
    }

    public string ParentMailboxName {
      get { return mailbox.SuperiorName; }
    }

    public string MailboxSeparator {
      get { return mailbox.HierarchyDelimiter; }
    }

    public long UidValidity {
      get { return mailbox.UidValidity; }
    }

    public long UnseenMessageCount {
      get { return mailbox.UnseenMessage; }
    }

    public long ExistMessageCount {
      get { return mailbox.ExistsMessage; }
    }

    public long RecentMessageCount {
      get { return mailbox.RecentMessage; }
    }

    public long FirstUnseenMessageNumber {
      get { return mailbox.FirstUnseen; }
    }

    public long NextUid {
      get { return mailbox.UidNext; }
    }

    public IImapMailboxFlagSet Flags {
      get { return mailbox.Flags; }
    }

    [CLSCompliant(false)]
    public ulong HighestModSeq {
      get { return mailbox.HighestModSeq; }
    }

    public bool ModificationSequences {
      get { return mailbox.ModificationSequences; }
    }

    public bool Exists {
      get { return !deleted && !mailbox.Flags.Has(ImapMailboxFlag.NonExistent); }
    }

    public bool IsUnselectable {
      get { return mailbox.IsUnselectable; }
    }

    public bool IsNonHierarchical {
      get { return mailbox.IsNonHierarchical; }
    }

    public bool CanCreateChild {
      get { return !mailbox.Flags.Has(ImapMailboxFlag.NoInferiors); }
    }

    public IImapMessageFlagSet ApplicableFlags {
      get { return mailbox.ApplicableFlags; }
    }

    public IImapMessageFlagSet PermanentFlags {
      get { return mailbox.PermanentFlags; }
    }

    public bool IsAllowedToCreateKeywords {
      get { return mailbox.PermanentFlags.Has(ImapMessageFlag.AllowedCreateKeywords); }
    }

    public bool IsUidPersistent {
      get { return mailbox.UidPersistent; }
    }

    protected ImapClient Client {
      get { return client; }
    }

    internal protected ImapMailbox Mailbox {
      get { return mailbox; }
    }

    internal ImapMailboxInfo(ImapClient client, ImapMailbox mailbox)
    {
      this.client = client;
      this.mailbox = mailbox;
    }

    private ImapClient client;
    private ImapMailbox mailbox;
    private bool deleted = false;

    /*
     * operations
     */
    public virtual void Refresh()
    {
      if (!Exists || mailbox.IsUnselectable)
        return;

      ImapClient.ThrowIfError(client.Session.Status(mailbox,
                                                    GetStatusDataItem(client.ServerCapabilities)));
    }

    internal static ImapStatusDataItem GetStatusDataItem(ImapCapabilityList serverCapabilities)
    {
      if (serverCapabilities.Has(ImapCapability.CondStore))
        return ImapStatusDataItem.StandardAll + ImapStatusDataItem.HighestModSeq;
      else
        return ImapStatusDataItem.StandardAll;
    }

    /*
     * SELECT/EXAMINE
     */
    public ImapOpenedMailboxInfo Open()
    {
      return client.OpenMailbox(this);
    }

    public ImapOpenedMailboxInfo Open(bool readOnly)
    {
      return client.OpenMailbox(this, readOnly);
    }

    /*
     * DELETE
     */
    public void Delete()
    {
      Delete(false);
    }

    public void Delete(bool unsubscribe)
    {
      if (client.IsSelected(this))
        client.CloseMailbox();

      if (Exists)
        ImapClient.ThrowIfError(client.Session.Delete(mailbox));

      if (unsubscribe)
        ImapClient.ThrowIfError(client.Session.Unsubscribe(mailbox.Name)); // mailbox is detached already

      deleted = true;
    }

    /*
     * CREATE
     */
    public void Create()
    {
      Create(false);
    }

    public void Create(bool subscribe)
    {
      ImapClient.ThrowIfError(client.Session.Create(mailbox.Name));

      deleted = false;

      if (subscribe)
        ImapClient.ThrowIfError(client.Session.Subscribe(mailbox.Name));
    }

    public static ImapMailboxInfo Create(ImapClient client, string mailboxName)
    {
      return Create(client, mailboxName, false);
    }

    public static ImapMailboxInfo Create(ImapClient client, string mailboxName, bool subscribe)
    {
      if (client == null)
        throw new ArgumentNullException("client");

      return client.CreateMailbox(mailboxName, subscribe);
    }

    public ImapMailboxInfo CreateChild(string childMailboxName)
    {
      return CreateChild(childMailboxName, false);
    }

    public ImapMailboxInfo CreateChild(string childMailboxName, bool subscribe)
    {
      if (!CanCreateChild)
        throw new ImapProtocolViolationException("can't create child under this mailbox");

      return client.CreateMailbox(mailbox.GetChildName(childMailboxName),
                                  subscribe);
    }

    /*
     * RENAME
     */
    public void MoveTo(ImapMailboxInfo destinationMailbox)
    {
      MoveTo(destinationMailbox, false);
    }

    public void MoveTo(ImapMailboxInfo destinationMailbox, bool subscribe)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");

      MoveTo(destinationMailbox.GetChildName(mailbox.LeafName), subscribe);
    }

    public void MoveTo(string newMailboxName)
    {
      MoveTo(newMailboxName, false);
    }

    public void MoveTo(string newMailboxName, bool subscribe)
    {
      if (newMailboxName == null)
        throw new ArgumentNullException("newMailboxName");

      bool readOnly;
      var selected = client.IsSelected(this, out readOnly);

      var oldMailboxName = mailbox.Name;

      if (selected)
        client.CloseMailbox();

      ImapClient.ThrowIfError(client.Session.Rename(mailbox, newMailboxName));

      if (subscribe) {
        ImapClient.ThrowIfError(client.Session.Unsubscribe(oldMailboxName));

        foreach (var childMailbox in client.GetMailboxes(oldMailboxName,
                                                         mailbox.HierarchyDelimiter,
                                                         true,
                                                         mailbox.IsNonHierarchical,
                                                         ImapMailboxListOptions.SubscribedOnly)) {
          childMailbox.Unsubscribe(false);
        }

        Subscribe(true);
      }

      if (selected)
        client.OpenMailbox(this, readOnly);
    }

    /*
     * SUBSCRIBE/UNSUBSCRIBE
     */
    public void Subscribe()
    {
      Subscribe(false);
    }

    public void Subscribe(bool recursive)
    {
      ImapClient.ThrowIfError(client.Session.Subscribe(mailbox));

      if (!recursive)
        return;

      foreach (var childMailbox in GetMailboxes(ImapMailboxListOptions.Default)) {
        ImapClient.ThrowIfError(client.Session.Subscribe(childMailbox.mailbox));
      }
    }

    public void Unsubscribe()
    {
      Unsubscribe(false);
    }

    public void Unsubscribe(bool recursive)
    {
      ImapClient.ThrowIfError(client.Session.Unsubscribe(mailbox));

      if (!recursive)
        return;

      foreach (var childMailbox in GetMailboxes(ImapMailboxListOptions.SubscribedOnly)) {
        ImapClient.ThrowIfError(client.Session.Unsubscribe(childMailbox.mailbox));
      }
    }

    /*
     * LIST/LSUB
     */
    public ImapMailboxInfo GetParent()
    {
      return GetParent(ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetParent(ImapMailboxListOptions options)
    {
      if (string.Empty.Equals(mailbox.SuperiorName))
        return null;

      return client.GetMailbox(mailbox.SuperiorName, options);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes()
    {
      return GetMailboxes(ImapMailboxListOptions.Default);
    }

    public IEnumerable<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
    {
      return client.GetMailboxes(mailbox, options);
    }

    /// <remarks>This method throws InvalidOperationException if HierarchyDelimiter is null or empty.</remarks>
    public string GetChildName(string childName)
    {
      return mailbox.GetChildName(childName);
    }

    /*
     * APPEND
     */
    public void AppendMessage(Stream stream)
    {
      AppendMessage(new ImapAppendMessage(stream));
    }

    public void AppendMessage(Stream stream, DateTimeOffset internalDate)
    {
      AppendMessage(new ImapAppendMessage(stream, internalDate));
    }

    public void AppendMessage(Stream stream, IImapMessageFlagSet flags)
    {
      AppendMessage(new ImapAppendMessage(stream, flags));
    }

    public void AppendMessage(Stream stream, DateTimeOffset internalDate, IImapMessageFlagSet flags)
    {
      AppendMessage(new ImapAppendMessage(stream, internalDate, flags));
    }

    public void AppendMessage(IImapAppendMessage message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ImapClient.ThrowIfError(Client.Session.Append(message, Mailbox));
    }

    public void AppendMessages(IEnumerable<IImapAppendMessage> messages)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");

      if (Client.ServerCapabilities.Has(ImapCapability.MultiAppend)) {
        ImapClient.ThrowIfError(Client.Session.AppendMultiple(messages, Mailbox));
      }
      else {
        foreach (var message in messages) {
          ImapClient.ThrowIfError(Client.Session.Append(message, Mailbox));
        }
      }
    }

    /*
     * GETQUOTAROOT/GETQUOTA
     */
    public IEnumerable<ImapQuota> GetQuota()
    {
      if (!client.IsCapable(ImapCapability.Quota))
        yield break;

      Dictionary<string, ImapQuota[]> quotaRoots;

      ImapClient.ThrowIfError(client.Session.GetQuotaRoot(Mailbox, out quotaRoots));

      foreach (var pair in quotaRoots) {
        foreach (var quota in pair.Value) {
          yield return quota;
        }
      }
    }

    public ImapQuota GetQuota(string quotaRoot)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");

      if (!client.IsCapable(ImapCapability.Quota))
        return null; // as default

      ImapQuota quota;

      ImapClient.ThrowIfError(client.Session.GetQuota(quotaRoot, out quota));

      return quota;
    }

    public double GetQuotaUsage(string quotaRoot, string resourceName)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");
      if (string.IsNullOrEmpty(resourceName))
        throw new ArgumentException("must be non-empty value", "resourceName");

      var quota = GetQuota(quotaRoot);

      if (quota == null)
        return 0.0; // as default

      foreach (var resource in quota.Resources) {
        if (string.Equals(resource.Name, resourceName, StringComparison.InvariantCultureIgnoreCase)) {
          if (resource.Limit == 0L)
            return 0.0;
          else
            return (double)resource.Usage / (double)resource.Limit;
        }
      }

      throw new ImapProtocolViolationException(string.Format("no such resource name: '{0}'", resourceName));
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(Uri baseUrl)
    {
      throw new NotSupportedException();
    }

    public override string ToString()
    {
      return string.Format("{{ImapMailboxInfo: Authority='{0}', FullName='{1}', UidValidity={2}, NextUid={3}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           FullName,
                           UidValidity,
                           NextUid);
    }
  }
}
