// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

using Smdn.Collections;
using Smdn.Formats;
using Smdn.IO;
using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public class ImapMessageInfo : ImapMessageInfoBase, IImapMessageAttribute {
    public long Sequence {
      get; private set;
    }

    public long Uid {
      get; private set;
    }

    public Uri Url {
      get { return uriBuilder.Uri; }
    }

    /*
     * IImapMessageDynamicAttribute
     */
    private IImapMessageDynamicAttribute dynamicAttribute;

    internal IImapMessageDynamicAttribute DynamicAttribute {
      get { UpdateDynamicAttributes(false); return dynamicAttribute; }
      set { dynamicAttribute = value; }
    }

    public IImapMessageFlagSet Flags {
      get { return DynamicAttribute.Flags; }
    }

    [CLSCompliant(false)]
    public ulong ModSeq {
      get { return DynamicAttribute.ModSeq; }
    }

    /*
     * IImapMessageStaticAttribute
     */
    private IImapMessageStaticAttribute staticAttribute;

    internal IImapMessageStaticAttribute StaticAttribute {
      get { UpdateStaticAttributes(); return staticAttribute; }
      set { staticAttribute = value; }
    }

    public IImapBodyStructure BodyStructure {
      get { return StaticAttribute.BodyStructure; }
    }

    public ImapEnvelope Envelope {
      get { return StaticAttribute.Envelope; }
    }

    public DateTimeOffset? InternalDate {
      get { return StaticAttribute.InternalDate; }
    }

    public long Length {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.Rfc822Size {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.GetBinarySizeOf(string section)
    {
      return StaticAttribute.GetBinarySizeOf(section);
    }

    /*
     * miscellaneous getter properties
     */
    public bool IsMarkedAsDeleted {
      get { return Flags.Has(ImapMessageFlag.Deleted); }
    }

    private Lazy<string> envelopeSubject;

    public string EnvelopeSubject {
      get { return envelopeSubject.Value; }
    }

    private Lazy<DateTimeOffset?> envelopeDate;

    public DateTimeOffset? EnvelopeDate {
      get { return envelopeDate.Value; }
    }

    private ImapSequenceSet uidSet;

    protected override ImapSequenceSet SequenceOrUidSet {
      get
      {
        if (uidSet == null)
          uidSet = ImapSequenceSet.CreateUidSet(Uid);

        return uidSet;
      }
    }

    internal ImapMessageInfo(ImapClient client, ImapOpenedMailboxInfo mailbox, long uid, long sequence)
      : base(client, mailbox)
    {
      this.Uid = uid;
      this.Sequence = sequence;

      (this as IImapUrl).SetBaseUrl(mailbox.Url);

      InitializeLazy();
    }

    private ImapUriBuilder uriBuilder;

    private void InitializeLazy()
    {
      envelopeSubject = new Lazy<string>(delegate {
        try {
          return TextConvert.FromMimeEncodedStringNullable(StaticAttribute.Envelope.Subject);
        }
        catch (FormatException) {
          return StaticAttribute.Envelope.Subject;
        }
      });

      envelopeDate = new Lazy<DateTimeOffset?>(delegate {
        try {
          return DateTimeConvert.FromRFC822DateTimeOffsetStringNullable(StaticAttribute.Envelope.Date);
        }
        catch (FormatException) {
          return default(DateTimeOffset);
        }
      });
    }

    /*
     * operations
     */
    public void Refresh()
    {
      UpdateStaticAttributes();
      UpdateDynamicAttributes(true);
    }

    internal static readonly ImapFetchDataItem FetchStaticDataItem =
      ImapFetchDataItem.BodyStructure +
      ImapFetchDataItem.Envelope + 
      ImapFetchDataItem.InternalDate +
      ImapFetchDataItem.Rfc822Size;

    private void UpdateStaticAttributes()
    {
      if (staticAttribute != null)
        return;

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapMessageStaticAttribute[] staticAttr;

      ImapClient.ThrowIfError(Client.Session.Fetch(SequenceOrUidSet,
                                                   FetchStaticDataItem,
                                                   out staticAttr));

      if (0 < staticAttr.Length)
        staticAttribute = staticAttr[0];
    }

    internal static ImapFetchDataItem GetFetchDynamicDataItem(ImapCapabilityList serverCapabilities)
    {
      if (serverCapabilities.Has(ImapCapability.CondStore))
        return ImapFetchDataItem.Flags + ImapFetchDataItem.ModSeq;
      else
        return ImapFetchDataItem.Flags;
    }

    private void UpdateDynamicAttributes(bool refresh)
    {
      if (!refresh && dynamicAttribute != null)
        return;

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapMessageDynamicAttribute[] dynamicAttr;

      ImapClient.ThrowIfError(Client.Session.Fetch(SequenceOrUidSet,
                                                   GetFetchDynamicDataItem(Client.ServerCapabilities),
                                                   out dynamicAttr));

      if (0 < dynamicAttr.Length)
        dynamicAttribute = dynamicAttr[0];
    }

    public void ToggleFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      ToggleFlags(new ImapMessageFlagList(flags.Prepend(flag)));
    }

    public void ToggleKeywords(string keyword, params string[] keywords)
    {
      ToggleFlags(new ImapMessageFlagList(keywords.Prepend(keyword)));
    }

    public void ToggleFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      var addFlags = new ImapMessageFlagList();
      var removeFlags = new ImapMessageFlagList();

      foreach (var flag in flagsAndKeywords) {
        if (dynamicAttribute.Flags.Has(flag))
          removeFlags.Add(flag);
        else
          addFlags.Add(flag);
      }

      if (0 < removeFlags.Count)
        Store(ImapStoreDataItem.RemoveFlags(removeFlags));

      if (0 < addFlags.Count)
        Store(ImapStoreDataItem.AddFlags(addFlags));
    }

    /*
    public override void Store(ImapStoreDataItem storeDataItem)
    {
      // TODO: update dynamic attributes
      base.Store(storeDataItem);
    }
    */

    /*
     * Open*(), Read*()
     */
    private static string GetSection(IImapBodyStructure structure)
    {
      if (structure == null)
        throw new ArgumentNullException("structure");

      return structure.Section;
    }

    private static string GetEncoding(IImapBodyStructure structure, bool decode)
    {
      if (decode && structure is ImapSinglePartBodyStructure)
        return (structure as ImapSinglePartBodyStructure).Encoding;
      else
        return null;
    }

    public Stream OpenRead()
    {
      return OpenRead(string.Empty, null, null);
    }

    public Stream OpenRead(bool decode)
    {
      return OpenRead(string.Empty, null, GetEncoding(BodyStructure, decode));
    }

    public Stream OpenRead(IImapBodyStructure section)
    {
      return OpenRead(GetSection(section), null, null);
    }

    public Stream OpenRead(IImapBodyStructure section, bool decode)
    {
      return OpenRead(GetSection(section), null, GetEncoding(BodyStructure, decode));
    }

    public Stream OpenRead(string section)
    {
      return OpenRead(section, null, null);
    }

    public Stream OpenRead(string section, ImapPartialRange range)
    {
      return OpenRead(section, range, null);
    }

    private Stream OpenRead(string section, ImapPartialRange range, string encoding)
    {
      //var decode = (encoding != null);

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      var stream = new ImapFetchMessageBodyStream(Client.Session,
                                                  true /*TODO: peek*/,
                                                  SequenceOrUidSet,
                                                  section,
                                                  range,
                                                  ImapFetchMessageBodyStream.DefaultFetchBlockSize);

      IImapMessageAttribute discard;

      ImapClient.ThrowIfError(stream.Prepare(null, out discard));

      return stream;
    }

    public TOutput ReadAs<TOutput>(Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(), converter);
    }

    public TOutput ReadAs<TOutput>(bool decode, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(decode), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TOutput ReadAs<TOutput>(IImapBodyStructure section, bool decode, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, decode), converter);
    }

    public TOutput ReadAs<TOutput>(string section, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TOutput ReadAs<TOutput>(string section, ImapPartialRange range, Converter<Stream, TOutput> converter)
    {
      return ReadAsCore(OpenRead(section, range), converter);
    }

    private TOutput ReadAsCore<TOutput>(Stream stream, Converter<Stream, TOutput> converter)
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      using (stream) {
        return converter(stream);
      }
    }

    public StreamReader OpenText()
    {
      return OpenText(string.Empty, Encoding.Default);
    }

    public StreamReader OpenText(string section)
    {
      return OpenText(section, Encoding.Default);
    }

    public StreamReader OpenText(Encoding encoding)
    {
      return OpenText(string.Empty, encoding);
    }

    public StreamReader OpenText(string section, Encoding encoding)
    {
      if (encoding == null)
        throw new ArgumentNullException("encoding");

      throw new NotImplementedException();
      //return new StreamReader(OpenRead(section, true), encoding);
    }

    public byte[] ReadAllBytes()
    {
      return ReadAllBytes(string.Empty, false);
    }

    public byte[] ReadAllBytes(IImapBodyStructure section)
    {
      return ReadAllBytes(GetSection(section), false);
    }

    public byte[] ReadAllBytes(IImapBodyStructure section, bool decode)
    {
      return ReadAllBytes(GetSection(section), decode);
    }

    public byte[] ReadAllBytes(string section, bool decode)
    {
      using (var stream = OpenRead(section, null, null)) {
        return stream.ReadToEnd();
      }
    }

    public IEnumerable<string> ReadLines(string section)
    {
      throw new NotImplementedException();
    }

    public string[] ReadAllLines(string section)
    {
      return ReadLines(section).ToArray();
    }

    public string ReadAllText(string section)
    {
      throw new NotImplementedException();
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(Uri baseUrl)
    {
      uriBuilder = new ImapUriBuilder(baseUrl);

      uriBuilder.Uid = Uid;
      uriBuilder.UidValidity = UidValidity;
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessageInfo: Authority='{0}', Mailbox='{1}', Sequence={2}, Uid={3}, UidValidity={4}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           Mailbox.FullName,
                           Sequence,
                           Uid,
                           UidValidity);
    }
  }
}
