// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public abstract class ImapMessageInfoBase {
    protected ImapClient Client {
      get; private set;
    }

    public ImapOpenedMailboxInfo Mailbox {
      get; private set;
    }

    public long UidValidity {
      get; protected set;
    }

    protected abstract ImapSequenceSet SequenceOrUidSet { get; }

    protected ImapMessageInfoBase(ImapClient client, ImapOpenedMailboxInfo mailbox)
    {
      if (client == null)
        throw new ArgumentNullException("client");
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");

      this.Client = client;
      this.Mailbox = mailbox;
      this.UidValidity = mailbox.UidValidity; // save current UIDVALIDITY
    }

    /*
     * operations
     */
    public void AddFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.AddFlags(flag, flags));
    }

    public void AddKeywords(string keyword, params string[] keywords)
    {
      Store(ImapStoreDataItem.AddFlags(keyword, keywords));
    }

    public void AddFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.AddFlags(flagsAndKeywords));
    }

    public void RemoveFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.RemoveFlags(flag, flags));
    }

    public void RemoveKeywords(string keyword, params string[] keywords)
    {
      Store(ImapStoreDataItem.RemoveFlags(keyword, keywords));
    }

    public void RemoveFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.RemoveFlags(flagsAndKeywords));
    }

    public void ReplaceFlags(ImapMessageFlag flag, params ImapMessageFlag[] flags)
    {
      Store(ImapStoreDataItem.ReplaceFlags(flag, flags));
    }

    public void ReplaceKeywords(string keyword, params string[] keywords)
    {
      Store(ImapStoreDataItem.ReplaceFlags(keyword, keywords));
    }

    public void ReplaceFlags(IImapMessageFlagSet flagsAndKeywords)
    {
      Store(ImapStoreDataItem.ReplaceFlags(flagsAndKeywords));
    }

    public virtual void Store(ImapStoreDataItem storeDataItem)
    {
      if (storeDataItem == null)
        throw new ArgumentNullException("storeDataItem");

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapClient.ThrowIfError(Client.Session.Store(SequenceOrUidSet, storeDataItem));
    }

    private static ImapStoreDataItem markAsDeletedStoreDataItem = ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted);

    public void MarkAsDeleted()
    {
      Store(markAsDeletedStoreDataItem);
    }

    public void Delete()
    {
      MarkAsDeleted();

      if (SequenceOrUidSet.IsUidSet && Client.ServerCapabilities.Has(ImapCapability.UidPlus))
        ImapClient.ThrowIfError(Client.Session.UidExpunge(SequenceOrUidSet));
      else
        ImapClient.ThrowIfError(Client.Session.Expunge());
    }

    public void CopyTo(ImapMailboxInfo destinationMailbox)
    {
      if (destinationMailbox == null)
        throw new ArgumentNullException("destinationMailbox");
      if (!destinationMailbox.Exists)
        throw new ImapProtocolViolationException("destination mailbox is not existent");

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapClient.ThrowIfError(Client.Session.Copy(SequenceOrUidSet, destinationMailbox.Mailbox));
    }

    public void MoveTo(ImapMailboxInfo destinationMailbox)
    {
      CopyTo(destinationMailbox);

      MarkAsDeleted();
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo CopyTo(string destinationMailboxName)
    {
      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, SequenceOrUidSet);

      ImapMailbox createdMailbox;

      ImapClient.ThrowIfError(Client.Session.Copy(SequenceOrUidSet,
                                                  ImapClient.GetValidatedMailboxName(destinationMailboxName),
                                                  out createdMailbox));

      if (createdMailbox == null)
        return null;
      else
        return new ImapMailboxInfo(Client, createdMailbox);
    }

    /// <returns>The mailbox if created, otherwise null.</returns>
    public ImapMailboxInfo MoveTo(string destinationMailboxName)
    {
      var ret = CopyTo(destinationMailboxName);

      MarkAsDeleted();

      return ret;
    }
  }
}
