// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Collections;
using Smdn.Net.Imap4;

namespace Smdn.Net.Imap4.Client {
  public class ImapMessageInfoList : ImapMessageInfoBase, IEnumerable<ImapMessageInfo> {
    protected override ImapSequenceSet SequenceOrUidSet {
      get { return sequenceOrUidSet; }
    }

    internal static ImapMessageInfoList CreateEmpty(ImapClient client,
                                                    ImapOpenedMailboxInfo mailbox,
                                                    ImapMessageFetchOptions fetchOptions,
                                                    bool orderBySequenceNumber)
    {
      return Create(client,
                    mailbox,
                    ImapSequenceSet.CreateSet(new long[0]),
                    fetchOptions,
                    orderBySequenceNumber);
    }

    internal static ImapMessageInfoList Create(ImapClient client,
                                               ImapOpenedMailboxInfo mailbox,
                                               ImapSequenceSet sequenceOrUidSet,
                                               ImapMessageFetchOptions fetchOptions,
                                               bool orderBySequenceNumber)
    {
      return new ImapMessageInfoList(client,
                                     mailbox,
                                     sequenceOrUidSet,
                                     fetchOptions,
                                     orderBySequenceNumber);
    }

    private ImapMessageInfoList(ImapClient client,
                                ImapOpenedMailboxInfo mailbox,
                                ImapSequenceSet sequenceOrUidSet,
                                ImapMessageFetchOptions fetchOptions,
                                bool orderBySequenceNumber)
      : base(client, mailbox)
    {
      this.sequenceOrUidSet = sequenceOrUidSet;
      this.fetchOptions = fetchOptions;
      this.orderBySequenceNumber = orderBySequenceNumber;
    }

    public IEnumerator<ImapMessageInfo> GetEnumerator()
    {
      if (sequenceOrUidSet.IsEmpty)
        yield break;

      Mailbox.CheckSelected();
      Mailbox.CheckUidValidity(UidValidity, sequenceOrUidSet);

      ImapFetchDataItem fetchDataItem;
      bool fetchStaticAttr, fetchDynamicAttr;

      TranslateFetchOption(Client, fetchOptions, out fetchDataItem, out fetchStaticAttr, out fetchDynamicAttr);

      foreach (var fetchSet in GetBatchedSequenceSet(sequenceOrUidSet)) {
        ImapMessageAttribute[] messages;

        ImapClient.ThrowIfError(Client.Session.Fetch(fetchSet,
                                                     fetchDataItem,
                                                     out messages));

        if (orderBySequenceNumber) {
          foreach (var sequence in fetchSet) {
            var message = messages.Find(delegate(ImapMessageAttribute attr) {
              return attr.Sequence == sequence;
            });

            yield return ToMessageInfo(message, fetchStaticAttr, fetchDynamicAttr);
          }
        }
        else {
          foreach (var message in messages) {
            yield return ToMessageInfo(message, fetchStaticAttr, fetchDynamicAttr);
          }
        }
      }
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return GetEnumerator();
    }

    private static void TranslateFetchOption(ImapClient client,
                                             ImapMessageFetchOptions options,
                                             out ImapFetchDataItem fetchDataItem,
                                             out bool fetchStaticAttr,
                                             out bool fetchDynamicAttr)
    {
      fetchStaticAttr = (int)(options & ImapMessageFetchOptions.FetchStaticAttribute) != 0;
      fetchDynamicAttr = (int)(options & ImapMessageFetchOptions.FetchDynamicAttribute) != 0;
      fetchDataItem = ImapFetchDataItem.Uid;

      if (fetchStaticAttr)
        fetchDataItem += ImapMessageInfo.FetchStaticDataItem;
      if (fetchDynamicAttr)
        fetchDataItem += ImapMessageInfo.GetFetchDynamicDataItem(client.ServerCapabilities);
    }

    private const long batchCount = 100L;

    private static IEnumerable<ImapSequenceSet> GetBatchedSequenceSet(ImapSequenceSet baseSequenceOrUidSet)
    {
      var baseMatchedResult = baseSequenceOrUidSet as ImapMatchedSequenceSet;

      if (baseMatchedResult != null && baseMatchedResult.IsSavedResult) {
        yield return baseMatchedResult;
        yield break;
      }

      var batched = new List<long>();

      foreach (var num in baseSequenceOrUidSet) {
        batched.Add(num);

        if (batchCount <= batched.Count) {
          yield return ImapSequenceSet.CreateSet(baseSequenceOrUidSet.IsUidSet, batched.ToArray());

          batched.Clear();
        }
      }

      if (0 < batched.Count)
        yield return ImapSequenceSet.CreateSet(baseSequenceOrUidSet.IsUidSet, batched.ToArray());
    }

    private ImapMessageInfo ToMessageInfo(ImapMessageAttribute message,
                                          bool fetchedStaticAttr,
                                          bool fetchedDynamicAttr)
    {
      var info = new ImapMessageInfo(Client, Mailbox, message.Uid, message.Sequence);

      if (fetchedStaticAttr)
        info.StaticAttribute = message.GetStaticAttributeImpl();
      if (fetchedDynamicAttr)
        info.DynamicAttribute = message.GetDynamicAttributeImpl();

      return info;
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessageInfoList: Mailbox='{0}', UidValidity={1}, SequenceOrUidSet={2}}}",
                           Mailbox.FullName,
                           UidValidity,
                           sequenceOrUidSet);
    }

    private ImapSequenceSet sequenceOrUidSet;
    private readonly ImapMessageFetchOptions fetchOptions;
    private readonly bool orderBySequenceNumber;
  }
}
