// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;

using Smdn.Collections;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  public class ImapOpenedMailboxInfo : ImapMailboxInfo, IDisposable {
    internal ImapOpenedMailboxInfo(ImapClient client, ImapMailbox mailbox)
      : base(client, mailbox)
    {
    }

    void IDisposable.Dispose()
    {
      this.Close();
    }

    public void Close()
    {
      Client.CloseMailbox();
    }

    internal void CheckSelected()
    {
      Client.ThrowIfNotSelected(this);
    }

    internal void CheckUidValidity(long uidValidity, ImapSequenceSet sequenceOrUidSet)
    {
      // TODO: impl
#if false
      if (!sequenceOrUidSet.IsUidSet)
        return;
      if (uidValidity != UidValidity)
        throw new ImapException("UIDVALIDITY value has been changed");
#endif
    }

    /*
     * operations
     */
    public override void Refresh()
    {
      CheckSelected();

      ImapClient.ThrowIfError(Client.Session.NoOp());
    }

    public void Expunge()
    {
      CheckSelected();

      ImapClient.ThrowIfError(Client.Session.Expunge());
    }

    /*
     * GetMessages by FETCHing
     */
    public ImapMessageInfoList GetMessages()
    {
      return GetMessages(ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList GetMessages(ImapMessageFetchOptions options)
    {
      Refresh();

      if (Mailbox.ExistsMessage <= 0L)
        return ImapMessageInfoList.CreateEmpty(Client,
                                               this,
                                               options,
                                               false);
      else
        return ImapMessageInfoList.Create(Client,
                                          this,
                                          ImapSequenceSet.CreateRangeSet(1L, Mailbox.ExistsMessage),
                                          options,
                                          false);
    }

    /*
     * GetMessage by FETCHing with uid or sequence
     */
    public ImapMessageInfo GetMessageByUid(long uid)
    {
      return GetMessageByUid(uid, ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfo GetMessageByUid(long uid, ImapMessageFetchOptions options)
    {
      /*
       * 6.4.8. UID Command
       *       A non-existent unique identifier is ignored without any error
       *       message generated.  Thus, it is possible for a UID FETCH command
       *       to return an OK without any data or a UID COPY or UID STORE to
       *       return an OK without performing any operations.
       */
      return GetMessage(ImapSequenceSet.CreateUidSet(uid), options);
    }

    public ImapMessageInfo GetMessageBySequence(long sequence)
    {
      return GetMessageBySequence(sequence, ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfo GetMessageBySequence(long sequence, ImapMessageFetchOptions options)
    {
      // Refresh();

      if (Mailbox.ExistsMessage < sequence)
        throw new ImapProtocolViolationException(string.Format("specified sequence number is greater than exist message count. (exist message count = {0})",
                                                               Mailbox.ExistsMessage));

      return GetMessage(ImapSequenceSet.CreateSet(sequence), options);
    }

    private ImapMessageInfo GetMessage(ImapSequenceSet sequenceOrUidSet, ImapMessageFetchOptions options)
    {
      var message = ImapMessageInfoList.Create(Client,
                                               this,
                                               sequenceOrUidSet,
                                               options,
                                               false).FirstOrDefault();

      if (message == null)
        throw new ImapMessageNotFoundException(sequenceOrUidSet);

      return message;
    }

    /*
     * GetMessages by SEARCHing
     */
    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria)
    {
      return GetMessages(searchCriteria, null, ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria,
                                           ImapMessageFetchOptions options)
    {
      return GetMessages(searchCriteria, null, options);
    }

    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria,
                                           Encoding encoding)
    {
      return GetMessages(searchCriteria, encoding, ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria,
                                           Encoding encoding,
                                           ImapMessageFetchOptions options)
    {
      if (searchCriteria == null)
        throw new ArgumentNullException("searchCriteria");

      CheckSelected();

      ImapMatchedSequenceSet matchedSequenceNumbers;

#if false
      if (Client.IsCapable(ImapCapability.Searchres))
        ImapClient.ThrowIfError(Client.Session.ESearch(searchCriteria,
                                                       encoding,
                                                       ImapSearchResultOptions.Save,
                                                       out matchedSequenceNumbers));
      else
#endif
        ImapClient.ThrowIfError(Client.Session.Search(searchCriteria,
                                                      encoding,
                                                      out matchedSequenceNumbers));

      return ImapMessageInfoList.Create(Client,
                                        this,
                                        matchedSequenceNumbers,
                                        options,
                                        false);
    }

    /*
     * GetMessages by SORTing
     */
    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria)
    {
      return GetSortedMessages(sortOrder, searchCriteria, null, ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria,
                                                 ImapMessageFetchOptions options)
    {
      return GetSortedMessages(sortOrder, searchCriteria, null, options);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria,
                                                 Encoding encoding)
    {
      return GetSortedMessages(sortOrder, searchCriteria, encoding, ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria,
                                                 Encoding encoding,
                                                 ImapMessageFetchOptions options)
    {
      if (sortOrder == null)
        throw new ArgumentNullException("sortOrder");
      if (searchCriteria == null)
        throw new ArgumentNullException("searchCriteria");

      Client.ThrowIfIncapable(ImapCapability.Sort);

      CheckSelected();

      ImapMatchedSequenceSet matchedSequenceNumbers;

      ImapClient.ThrowIfError(Client.Session.Sort(sortOrder,
                                                  searchCriteria,
                                                  encoding,
                                                  out matchedSequenceNumbers));

      return ImapMessageInfoList.Create(Client,
                                        this,
                                        matchedSequenceNumbers,
                                        options,
                                        true);
    }

    /*
     * miscellaneous operation methods
     */

    /*
     * WaitForMessageArrival
     */
    private const int defaultPollingInterval = 10 * 60 * 1000; // 10 mins
    private const int maximumPollingInterval = 30 * 60 * 1000; // 30 mins

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan timeout)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   (int)timeout.TotalMilliseconds,
                                   ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan timeout,
                                                     ImapMessageFetchOptions options)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   (int)timeout.TotalMilliseconds,
                                   options);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsTimeout)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   millisecondsTimeout,
                                   ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsTimeout,
                                                     ImapMessageFetchOptions options)
    {
      return WaitForMessageArrival(defaultPollingInterval,
                                   millisecondsTimeout,
                                   options);
    }

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan pollingInterval,
                                                     TimeSpan timeout)
    {
      return WaitForMessageArrival((int)pollingInterval.TotalMilliseconds,
                                   (int)timeout.TotalMilliseconds,
                                   ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(TimeSpan pollingInterval,
                                                     TimeSpan timeout,
                                                     ImapMessageFetchOptions options)
    {
      return WaitForMessageArrival((int)pollingInterval.TotalMilliseconds,
                                   (int)timeout.TotalMilliseconds,
                                   options);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsPollingInterval,
                                                     int millisecondsTimeout)
    {
      return WaitForMessageArrival(millisecondsPollingInterval,
                                   millisecondsTimeout,
                                   ImapMessageFetchOptions.Default);
    }

    public ImapMessageInfoList WaitForMessageArrival(int millisecondsPollingInterval,
                                                     int millisecondsTimeout,
                                                     ImapMessageFetchOptions options)
    {
      if (millisecondsTimeout < -1)
        throw new ArgumentOutOfRangeException("millisecondsTimeout", "must be greater than or equals to -1");

      var asyncResult = BeginWaitForMessageArrival(millisecondsPollingInterval, options, null, null);

      asyncResult.AsyncWaitHandle.WaitOne(millisecondsTimeout);

      return EndWaitForMessageArrival(asyncResult);
    }

    /*
     * BeginWaitForMessageArrival
     */
    public IAsyncResult BeginWaitForMessageArrival(object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival(defaultPollingInterval,
                                        ImapMessageFetchOptions.Default,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(TimeSpan pollingInterval,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival((int)pollingInterval.Milliseconds,
                                        ImapMessageFetchOptions.Default,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(int millisecondsPollingInterval,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival(millisecondsPollingInterval,
                                        ImapMessageFetchOptions.Default,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(TimeSpan pollingInterval,
                                                   ImapMessageFetchOptions options,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      return BeginWaitForMessageArrival((int)pollingInterval.TotalMilliseconds,
                                        options,
                                        asyncState,
                                        asyncCallback);
    }

    public IAsyncResult BeginWaitForMessageArrival(int millisecondsPollingInterval,
                                                   ImapMessageFetchOptions options,
                                                   object asyncState,
                                                   AsyncCallback asyncCallback)
    {
      if (millisecondsPollingInterval <= 0)
        throw new ArgumentOutOfRangeException("millisecondsPollingInterval", "must be non-zero positive number");
      if (waitForMessageArrivalAsyncResult != null)
        throw new InvalidOperationException("another operation proceeding");

      var isIdleCapable = Client.ServerCapabilities.Has(ImapCapability.Idle);

      if (!isIdleCapable && maximumPollingInterval <= millisecondsPollingInterval)
        throw new ArgumentOutOfRangeException("millisecondsPollingInterval",
                                              string.Format("maximum interval is {0} but requested was {1}", maximumPollingInterval, millisecondsPollingInterval));

      var proc = isIdleCapable
        ? new WaitForMessageArrivalProc(WaitForMessageArrivalIdle)
        : new WaitForMessageArrivalProc(WaitForMessageArrivalNoOp);

      waitForMessageArrivalAsyncResult = new WaitForMessageArrivalAsyncResult(millisecondsPollingInterval,
                                                                              Mailbox,
                                                                              options,
                                                                              proc,
                                                                              asyncState,
                                                                              asyncCallback);

      waitForMessageArrivalAsyncResult.BeginProc();

      return waitForMessageArrivalAsyncResult;
    }

    public ImapMessageInfoList EndWaitForMessageArrival(IAsyncResult asyncResult)
    {
      if (asyncResult != waitForMessageArrivalAsyncResult)
        throw new ArgumentException("invalid IAsyncResult", "asyncResult");

      try {
        waitForMessageArrivalAsyncResult.EndProc();

        return GetArrivalMessages(waitForMessageArrivalAsyncResult);
      }
      finally {
        waitForMessageArrivalAsyncResult = null;
      }
    }

    private ImapMessageInfoList GetArrivalMessages(WaitForMessageArrivalContext context)
    {
      try {
        if (!context.MessageArrived ||
            context.Mailbox.ExistsMessage <= context.CurrentMessageCount)
          return ImapMessageInfoList.CreateEmpty(Client,
                                                 this,
                                                 context.FetchOptions,
                                                 false);

        return ImapMessageInfoList.Create(Client,
                                          this,
                                          ImapSequenceSet.CreateRangeSet(context.CurrentMessageCount + 1,
                                                                         context.Mailbox.ExistsMessage),
                                          context.FetchOptions,
                                          false);
      }
      finally {
        context.Dispose();
      }
    }

    private delegate bool WaitForMessageArrivalProc(WaitForMessageArrivalContext context);
    private WaitForMessageArrivalAsyncResult waitForMessageArrivalAsyncResult = null;

    private class WaitForMessageArrivalContext : IDisposable {
      public bool MessageArrived {
        get; protected set;
      }

      public WaitHandle AbortEvent {
        get { return abortEvent; }
      }

      public long CurrentMessageCount {
        get; set;
      }

      public int PollingInterval {
        get; private set;
      }

      public ImapMailbox Mailbox {
        get; private set;
      }

      public ImapMessageFetchOptions FetchOptions {
        get; private set;
      }

      public WaitForMessageArrivalContext(int pollingInterval,
                                          ImapMailbox mailbox,
                                          ImapMessageFetchOptions fetchOptions)
      {
        this.PollingInterval = pollingInterval;
        this.Mailbox = mailbox;
        this.CurrentMessageCount = mailbox.ExistsMessage;
        this.FetchOptions = fetchOptions;
      }

      public void Dispose()
      {
        if (abortEvent != null) {
          abortEvent.Close();
          abortEvent = null;
        }
      }

      protected void Abort()
      {
        abortEvent.Set();
      }

      private ManualResetEvent abortEvent = new ManualResetEvent(false);
    }

    private class WaitForMessageArrivalAsyncResult :
      WaitForMessageArrivalContext,
      IAsyncResult
    {
      public object AsyncState {
        get; private set;
      }

      public WaitHandle AsyncWaitHandle {
        get { return procAsyncResult.AsyncWaitHandle; }
      }

      public bool CompletedSynchronously {
        get { return false; }
      }

      public bool IsCompleted {
        get { return procAsyncResult.AsyncWaitHandle.WaitOne(0); }
      }

      private AsyncCallback asyncCallback;
      private WaitForMessageArrivalProc proc;
      private IAsyncResult procAsyncResult;

      public WaitForMessageArrivalAsyncResult(int pollingInterval,
                                              ImapMailbox mailbox,
                                              ImapMessageFetchOptions fetchOptions,
                                              WaitForMessageArrivalProc proc,
                                              object asyncState,
                                              AsyncCallback asyncCallback)
        : base(pollingInterval, mailbox, fetchOptions)
      {
        this.proc = proc;
        this.AsyncState = asyncState;
        this.asyncCallback = asyncCallback;
      }

      public void BeginProc()
      {
        procAsyncResult = proc.BeginInvoke(this, null, null);
      }

      public void EndProc()
      {
        Abort();

        MessageArrived = proc.EndInvoke(procAsyncResult);

        if (MessageArrived && asyncCallback != null)
          asyncCallback(this);
      }
    }

    private bool WaitForMessageArrivalIdle(WaitForMessageArrivalContext context)
    {
      var waitHandles = new WaitHandle[] {context.AbortEvent, null};

      for (;;) {
        if (context.AbortEvent.WaitOne(0))
          return false; // aborted

        var idleAsyncResult = Client.Session.BeginIdle(context, delegate(object idleState, ImapUpdatedStatus updatedStatus) {
          var ctx = idleState as WaitForMessageArrivalAsyncResult;

          ctx.CurrentMessageCount -= updatedStatus.Expunged.Length;

          if (updatedStatus.Exists.HasValue && ctx.CurrentMessageCount < updatedStatus.Exists.Value)
            return false; // DONE
          else
            return true; // keep IDLEing
        });

        waitHandles[1] = idleAsyncResult.AsyncWaitHandle;

        var index = WaitHandle.WaitAny(waitHandles, 180 * 1000); // send DONE 3 minute after IDLE started

        Client.Session.EndIdle(idleAsyncResult);

        switch (index) {
          case 0: // aborted
            return false;

          case 1: // IDLE completed
            return true;

          case WaitHandle.WaitTimeout:
          default:
            continue; // re-IDLE
        }
      }
    }

    private bool WaitForMessageArrivalNoOp(WaitForMessageArrivalContext context)
    {
      for (;;) {
        if (context.AbortEvent.WaitOne(context.PollingInterval))
          return false; // aborted

        var result = ImapClient.ThrowIfError(Client.Session.NoOp());

        foreach (var response in result.ReceivedResponses) {
          var dataResponse = response as ImapDataResponse;

          if (dataResponse == null)
            continue;

          if (dataResponse.Type == ImapDataResponseType.Expunge)
            context.CurrentMessageCount--;
        }

        if (context.CurrentMessageCount < context.Mailbox.ExistsMessage)
          return true; // new message arrival
      }
    }
  }
}
