// 
// Author:
//       smdn <smdn@mail.invisiblefulmoon.net>
// 
// Copyright (c) 2008-2010 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.IO;
using System.Collections.Generic;

using Smdn.Formats;
using Smdn.Net.MessageAccessProtocols;

namespace Smdn.Net.Imap4.Protocol {
  public abstract class ImapSender {
    internal protected int UnsentFragments {
      get { return fragments.Count; }
    }

    protected LineOrientedBufferedStream Stream {
      get { return stream; }
    }

    protected ImapSender(LineOrientedBufferedStream stream)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");

      this.stream = stream;
    }

    public void Send()
    {
      for (;;) {
        if (fragments.Count == 0)
          return;

        var fragment = fragments.Dequeue();

        if (fragment is BytesFragment)
          stream.WriteToCrLf((fragment as BytesFragment).Bytes);
        else if (fragment is StreamFragment)
          stream.Write((fragment as StreamFragment).Stream);
        else if (fragment is WaitContinuationFragment)
          return;
      }
    }

    protected void ClearUnsent()
    {
      fragments.Clear();
    }

    protected void Enqueue(byte[] bytes)
    {
      fragments.Enqueue(new BytesFragment(bytes));
    }

    protected void Enqueue(IEnumerable<ImapString> args)
    {
      var delim = false;

      foreach (var arg in args) {
        if (delim)
          fragments.Enqueue(BytesFragment.Delimiter); // SP
        else
          delim = true;

        var str = (arg is ImapCombinableDataItem) ? (arg as ImapCombinableDataItem).GetCombined() : arg;

        if (str is ImapStringList) {
          var parenthesize = (str is ImapParenthesizedString);

          if (parenthesize)
            fragments.Enqueue(BytesFragment.OpenParen);

          Enqueue(str as ImapStringList);

          if (parenthesize)
            fragments.Enqueue(BytesFragment.CloseParen);
        }
        else {
          if (str is ImapQuotedString) {
            // quoted          = DQUOTE *QUOTED-CHAR DQUOTE
            var quoted = str as ImapQuotedString;

            fragments.Enqueue(new BytesFragment(string.Format("\"{0}\"", quoted.Escaped)));
          }
          else if (str is IImapLiteralString) {
            /*
             * RFC 3501 - INTERNET MESSAGE ACCESS PROTOCOL - VERSION 4rev1
             * http://tools.ietf.org/html/rfc3501
             * 
             *    literal         = "{" number "}" CRLF *CHAR8
             *                        ; Number represents the number of CHAR8s
             * 
             * IMAP4 non-synchronizing literals
             * http://tools.ietf.org/html/rfc2088
             * 
             * RFC 3516 - IMAP4 Binary Content Extension
             * http://tools.ietf.org/html/rfc3516
             * 
             * RFC 4466 - Collected Extensions to IMAP4 ABNF
             * http://tools.ietf.org/html/rfc4466
             *
             *    literal8        = "~{" number ["+"] "}" CRLF *OCTET
             *                       ;; A string that might contain NULs.
             *                       ;; <number> represents the number of OCTETs
             *                       ;; in the response string.
             *                       ;; The "+" is only allowed when both LITERAL+ and
             *                       ;; BINARY extensions are supported by the server.
             */
            var literal = str as IImapLiteralString;
            var stream = literal.GetLiteralStream();

            var nonSync = (literal.Options & ImapLiteralOptions.SynchronizationMode) == ImapLiteralOptions.NonSynchronizing;
            var literal8 = (literal.Options & ImapLiteralOptions.LiteralMode) == ImapLiteralOptions.Literal8;

            fragments.Enqueue(new BytesFragment((literal8 ? "~{" : "{") +
                                                stream.Length.ToString() +
                                                (nonSync ? "+}\x0d\x0a" : "}\x0d\x0a")));

            if (!nonSync)
              fragments.Enqueue(new WaitContinuationFragment());

            // *CHAR8 / *OCTET
            fragments.Enqueue(new StreamFragment(stream));
          }
          else if (str is ImapPreformattedString) {
            fragments.Enqueue(new BytesFragment((str as ImapPreformattedString).Octets));
          }
          else {
            // atom            = 1*ATOM-CHAR
            fragments.Enqueue(new BytesFragment((string)str));
          }
        }
      }
    }

    private /*readonly*/ LineOrientedBufferedStream stream;
    private /*readonly*/ Queue<IFragment> fragments = new Queue<IFragment>();

    private interface IFragment {
    }

    private class WaitContinuationFragment : IFragment {
    }

    private class BytesFragment : IFragment {
      public byte[] Bytes;

      public BytesFragment(string str)
        : this(NetworkTransferEncoding.Transfer7Bit.GetBytes(str))
      {
      }

      public BytesFragment(byte[] bytes)
      {
        this.Bytes = bytes;
      }

      public static BytesFragment Delimiter = new BytesFragment(new[] {Octets.SP});
      public static BytesFragment OpenParen = new BytesFragment(new[] {ImapOctets.OpenParen});
      public static BytesFragment CloseParen = new BytesFragment(new[] {ImapOctets.CloseParen});
    }

    private class StreamFragment : IFragment {
      public Stream Stream;

      public StreamFragment(Stream stream)
      {
        this.Stream = stream;
      }
    }
  }
}